'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.TransformerOptions = exports.ParserOptions = exports.Script = undefined;

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _getPrototypeOf = require('babel-runtime/core-js/object/get-prototype-of');

var _getPrototypeOf2 = _interopRequireDefault(_getPrototypeOf);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _possibleConstructorReturn2 = require('babel-runtime/helpers/possibleConstructorReturn');

var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);

var _inherits2 = require('babel-runtime/helpers/inherits');

var _inherits3 = _interopRequireDefault(_inherits2);

var _index = require('../index');

var _index2 = _interopRequireDefault(_index);

var _asset = require('./asset');

var _asset2 = _interopRequireDefault(_asset);

var _docblockParser = require('docblock-parser');

var _docblockParser2 = _interopRequireDefault(_docblockParser);

var _babel = require('./parsers/babel');

var _indexer = require('./script/indexer');

var _indexer2 = _interopRequireDefault(_indexer);

var _util = require('../util');

var _get2 = require('lodash/get');

var _get3 = _interopRequireDefault(_get2);

var _result = require('lodash/result');

var _result2 = _interopRequireDefault(_result);

var _mapValues = require('lodash/mapValues');

var _mapValues2 = _interopRequireDefault(_mapValues);

var _isString = require('lodash/isString');

var _isString2 = _interopRequireDefault(_isString);

var _isError = require('lodash/isError');

var _isError2 = _interopRequireDefault(_isError);

var _isArray = require('lodash/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _attempt = require('lodash/attempt');

var _attempt2 = _interopRequireDefault(_attempt);

var _path = require('path');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var EXTENSIONS = ['js', 'jsx', 'cjsx', 'coffee', 'es6'];
var GLOB = '**/*.{' + EXTENSIONS.join(',') + '}';

/**
 * The Script Asset is a wrapper around es6 javascript files.
 *
 * Scripts are parsed using the babylon parser and compiled
 * with babel's transformer using the skypager preset.
 *
 * Script builds a metadata structure combining the information
 * we pull from the parser's AST, by parsing the docblocks,
 * as well as things such as where the asset lives
 * within the folder structure.
 *
 * @name Script
 * @extends Asset
 *
*/

var Script = exports.Script = (function (_Asset) {
  (0, _inherits3.default)(Script, _Asset);

  function Script(uri) {
    var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
    (0, _classCallCheck3.default)(this, Script);

    var _this = (0, _possibleConstructorReturn3.default)(this, (0, _getPrototypeOf2.default)(Script).call(this, uri, options));

    _this.transformer = function () {
      return (0, _babel.transformer)(_this.raw, _this);
    };
    _this.indexer = function () {
      return (0, _indexer2.default)(_this.parsed, _this);
    };

    _this.lazy('compiled', function () {
      return (0, _babel.compiler)(_this.readSync().raw, _this);
    });
    return _this;
  }

  (0, _createClass3.default)(Script, [{
    key: 'compileToDisk',
    value: function compileToDisk(outputPath) {
      var _this2 = this;

      outputPath = outputPath || this.compiledPath;

      return this.project.ensureFile(outputPath).then(function () {
        return _this2.project.outputFile(outputPath, _this2.compiled);
      }).then(function () {
        return outputPath;
      });
    }
  }, {
    key: 'compileToDiskSync',
    value: function compileToDiskSync(outputPath) {
      outputPath = outputPath || this.compiledPath;

      this.project.ensureFileSync(outputPath);
      this.project.outputFileSync(outputPath, this.compiled);

      return this;
    }
  }, {
    key: 'parser',
    value: function parser() {
      try {
        return (0, _babel.parser)(this.raw, this);
      } catch (error) {
        console.log('Parser Error', error);
      }
    }
  }, {
    key: 'getData',
    value: function getData() {
      if (!this.raw || this.raw.length === 0) {
        this.runImporter('disk', { asset: this, sync: true });
      }

      return this.metadata || {};
    }
  }, {
    key: 'getClassPropertyValue',
    value: function getClassPropertyValue(name) {
      if (this.indexed && this.indexes._classProperties[name]) {
        var prop = this.indexes._classProperties[name];

        return prop.value && prop.value.type === 'StringLiteral' ? prop.value.extra.rawValue : prop.value;
      }
    }
  }, {
    key: 'evalWithHelpers',

    /**
     * @todo replace with nodes vm module
     */
    value: function evalWithHelpers() {
      var helpers = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

      var project = this.project;
      var script = this;

      helpers = (0, _util.defaults)(helpers, {
        asset: script,
        project: project,
        script: script,
        skypager: _index2.default
      });

      var fn = (0, _util.noConflict)((0, _asyncToGenerator3.default)(_regenerator2.default.mark(function _callee() {
        return _regenerator2.default.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                try {
                  eval(script.compiled);
                } catch (error) {
                  project.error('error in evaluating script', error);
                }

              case 1:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      })), helpers);

      fn();
    }
  }, {
    key: 'generateId',
    value: function generateId() {
      var base = '' + this.paths.relative.replace(this.extension, '');

      if (base.match(/\/index$/)) {
        base = base.replace(/\/index$/, '');
      }

      return base;
    }
  }, {
    key: 'compiledPath',
    get: function get() {
      return this.paths.compiled || (0, _path.join)(this.collection.compiledPath, this.paths.relative);
    }
  }, {
    key: 'requiredFromCompiledSource',
    get: function get() {
      try {
        return require(this.compiledPath);
      } catch (error) {
        return false;
      }
    }
  }, {
    key: 'requiredFromSource',
    get: function get() {
      try {
        return require(this.paths.absolute);
      } catch (error) {
        return false;
      }
    }
  }, {
    key: 'requireStatus',
    get: function get() {
      this.readSync();

      var abs = this.paths.absolute;
      var compiled = this.compiledPath;

      return {
        raw: !(0, _isError2.default)((0, _attempt2.default)(function () {
          return require(abs);
        })),
        compiled: !(0, _isError2.default)((0, _attempt2.default)(function () {
          return require(compiled);
        })) && !this.project.readFileSync(compiled).toString().match(/Error while compiling/)
      };
    }
  }, {
    key: 'parserOptions',
    get: function get() {
      return ParserOptions;
    }
  }, {
    key: 'transformerOptions',
    get: function get() {
      return TransformerOptions;
    }
  }, {
    key: 'exports',
    get: function get() {
      return (0, _get3.default)(this.modulesMetadata, 'exports.exported', []);
    }
  }, {
    key: 'imports',
    get: function get() {
      return (0, _get3.default)(this.modulesMetadata, 'imports', []).map(function (i) {
        return i.source;
      });
    }
  }, {
    key: 'classPropertyNames',
    get: function get() {
      if (this.indexed) {
        return keys(this.indexes._classProperties);
      }
    }
  }, {
    key: 'reactDisplayName',
    get: function get() {
      if (this.isReactComponent) {
        return this.getClassPropertyValue('displayName');
      }
    }
  }, {
    key: 'hasReactPropTypes',
    get: function get() {
      return this.isReactComponent && this.classPropertyNames.indexOf('propTypes') >= 0 && this.getClassPropertyValue('propTypes').properties;
    }
  }, {
    key: 'reactPropTypesDeclaration',
    get: function get() {
      if (this.hasReactPropTypes) {
        return this.getClassPropertyValue('propTypes');
      }
    }
  }, {
    key: 'reactPropDocs',
    get: function get() {
      if (this.hasReactPropTypes) {
        return this.reactPropTypesDeclaration.properties.map(function (prop) {
          return prop.leadingComments;
        });
      }
    }
  }, {
    key: 'reactPropNames',
    get: function get() {
      if (this.hasReactPropTypes) {
        return this.getClassPropertyValue('propTypes').properties.map(function (node) {
          return node.key && node.key.name;
        });
      }
    }
  }, {
    key: 'isReactComponent',
    get: function get() {
      if (this.defaultExportType === 'ClassDeclaration') {
        var superClass = (0, _result2.default)(this, 'defaultExportNode.declaration.superClass.name');

        if (superClass && superClass.match(/Component$/)) {
          return true;
        }
      }
    }
  }, {
    key: 'exportedClassNames',
    get: function get() {
      if (this.indexed) {
        var exports = this.indexes._exportNodes;

        return exports.filter(function (node) {
          return (0, _get3.default)(node, 'declaration.type') === 'ClassDeclaration';
        }).filter(function (node) {
          return (0, _get3.default)(node, 'declaration.id.type') === 'Identifier';
        }).map(function (node) {
          return (0, _get3.default)(node, 'declaration.id.name');
        });
      }
    }
  }, {
    key: 'importsLocalModules',
    get: function get() {
      if (this.indexed) {
        return this.importNodes.filter(function (n) {
          return n.source && n.source.value.match(/^\./);
        }).map(function (n) {
          return n.source.value;
        });
      }
    }
  }, {
    key: 'importsModuleNames',
    get: function get() {
      if (this.indexed && this.importNodes) {
        return this.importNodes.map(function (n) {
          return n.source && n.source.value;
        });
      }
    }
  }, {
    key: 'importNodes',
    get: function get() {
      if (this.indexed) {
        return this.indexes._importNodes;
      }
    }
  }, {
    key: 'exportNodes',
    get: function get() {
      if (this.indexed) {
        return this.indexes._exportNodes;
      }
    }
  }, {
    key: 'hasDefaultDocblock',
    get: function get() {
      return this.defaultExportNode && this.defaultExportNode.leadingComments;
    }
  }, {
    key: 'defaultDocblock',
    get: function get() {
      if (this.indexed && this.hasDefaultDocblock) {
        var leadingComments = this.defaultExportNode && this.defaultExportNode.leadingComments;

        var blockContent = leadingComments.map(function (c) {
          return c.value;
        }).filter(function (v) {
          return v && v.length > 0;
        }).join("\n");

        return _docblockParser2.default.parse(blockContent);
      }
    }
  }, {
    key: 'defaultExportNode',
    get: function get() {
      if (this.indexed) {
        return this.exportNodes.find(function (d) {
          return d.type === 'ExportNamedDeclaration';
        });
      }
    }
  }, {
    key: 'defaultExportName',
    get: function get() {
      return (0, _result2.default)(this, 'defaultExportNode.declaration.id.name');
    }
  }, {
    key: 'defaultExportType',
    get: function get() {
      return (0, _result2.default)(this, 'defaultExportNode.declaration.type');
    }
  }, {
    key: 'modulesMetadata',
    get: function get() {
      if (this.transformerMetadata) {
        return this.transformerMetadata;
      } else {
        return {
          imports: [],
          exports: {
            exported: []
          }
        };
      }
    }
  }, {
    key: 'transformerMetadata',
    get: function get() {
      return (0, _get3.default)(this.transformed, 'metadata.modules');
    }

    /**
     * @alias modulesMetadata
     */

  }, {
    key: 'moduleMetadata',
    get: function get() {
      return this.modulesMetadata;
    }
  }, {
    key: 'programBody',
    get: function get() {
      return (0, _get3.default)(this.parsed, 'program.body');
    }
  }, {
    key: 'frontmatter',
    get: function get() {
      return this.metadata;
    }
  }, {
    key: 'docDescription',
    get: function get() {
      return this.get('defaultDocblock.text', '');
    }
  }, {
    key: 'tags',
    get: function get() {
      var tags = this.get('defaultDocblock.tags', {});

      return (0, _mapValues2.default)(tags || {}, function (v, k) {
        // coerce pluralized tags into arrays
        if ((0, _isString2.default)(v) && k.match(/s$/)) {
          return [v.trim()];
        }

        if ((0, _isString2.default)(v)) {
          return v.trim();
        }

        if ((0, _isArray2.default)(v)) {
          return v.map(function (val) {
            return (0, _isString2.default)(val) ? val.trim() : val;
          });
        }

        return v;
      });
    }
  }, {
    key: 'metadata',
    get: function get() {
      return (0, _extends3.default)({}, this.tags, {
        docDescription: this.docDescription,
        exports: this.exports,
        imports: this.imports,
        depth: this.depth,
        componentInfo: this.componentInfo,
        asset: this.assetMetadata
      });
    }
  }, {
    key: 'componentInfo',
    get: function get() {
      return this.isReactComponent ? { displayName: this.reactDisplayName, propNames: this.reactPropNames } : {};
    }
  }], [{
    key: 'decorateCollection',
    value: function decorateCollection(collection) {
      var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

      defineProperty(collection, 'compiledPath', {
        enumerable: false,
        configurable: true,
        get: function get() {
          return collection.project.paths.join('lib');
        }
      });
    }
  }]);
  return Script;
})(_asset2.default);

Script.EXTENSIONS = EXTENSIONS;
Script.GLOB = GLOB;
exports.default = Script;
var ParserOptions = exports.ParserOptions = {
  sourceType: 'module',
  plugins: ['jsx', 'asyncFunctions', 'classConstructorCall', 'doExpressions', 'trailingFunctionCommas', 'objectRestSpread', 'decorators', 'classProperties', 'exportExtensions', 'asyncGenerators', 'functionBind', 'functionSent']
};

var TransformerOptions = exports.TransformerOptions = {
  presets: [global.$skypager && global.$skypager.babelPresetSkypager || 'babel-preset-skypager']
};

var _Object = Object;
var defineProperty = _Object.defineProperty;
var keys = _Object.keys;