'use strict';

var _defineEnumerableProperties2 = require('babel-runtime/helpers/defineEnumerableProperties');

var _defineEnumerableProperties3 = _interopRequireDefault(_defineEnumerableProperties2);

var _keys = require('babel-runtime/core-js/object/keys');

var _keys2 = _interopRequireDefault(_keys);

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _assign = require('babel-runtime/core-js/object/assign');

var _assign2 = _interopRequireDefault(_assign);

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _index = require('./index');

var _index2 = _interopRequireDefault(_index);

var _util = require('./util');

var util = _interopRequireWildcard(_util);

var _lodash = require('lodash');

var _path = require('path');

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _pathExists = require('path-exists');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Fallbacks = {};

var Registry = (function () {
  (0, _createClass3.default)(Registry, null, [{
    key: 'build',
    value: function build(host, helper, options) {
      var registry = new Registry(host, helper, options);

      if (helper && helper.decorateRegistry) {
        helper.decorateRegistry(registry);
      }

      return registry;
    }
  }]);

  function Registry(host, helper) {
    var options = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];
    (0, _classCallCheck3.default)(this, Registry);

    if (process.env.NODE_ENV !== 'test') {
      (0, _invariant2.default)(host && host.type, 'Registries need to be created with a host that has a type property.');

      (0, _invariant2.default)(helper, 'Registries need to be created with a helper class that has a fromDefinition method on it');
    }

    this.name = options.name || util.tableize(helper.name || '');
    this.root = (0, _path.join)(options.root || host.root, this.name);

    var registry = {};
    var aliases = {};

    util.hide.getter(this, 'host', function () {
      return host;
    });
    util.hide.getter(this, 'aliases', function () {
      return aliases;
    });
    util.hide.getter(this, 'registry', function () {
      return registry;
    });
    util.hide.getter(this, 'helper', function () {
      return helper;
    });

    this.loaded = false;

    if (host.type === 'framework') {
      Fallbacks[this.name] = this;

      if ((0, _pathExists.sync)(this.indexScript = (0, _path.join)(this.root, 'index.js'))) {
        this.runLoader(require(this.indexScript));
      }
    }

    buildAtInterface(this);

    this.loaded = true;

    util.hide.property(this, '__helperContext', {
      registry: this,
      project: this.host.type === 'project' ? this.host : null
    });
  }

  /**
   * Remove a helper from this registry
   *
   * @param {Helper.id} helperId - the id the helper was registered with
  */

  (0, _createClass3.default)(Registry, [{
    key: 'remove',
    value: function remove(helperId) {
      var _this = this;

      var instance = this.lookup(helperId, false);
      if (instance) {
        if (instance.name) {
          delete (this.aliases, instance.name);
        }

        if (instance.aliases) {
          instance.aliases.forEach(function (a) {
            return delete _this.aliases[a];
          });
        }

        delete this.aliases[instance.id];
        delete this.registry[instance.id];
      }
    }

    /**
     * run a helper's main function.
     *
     * using the helperId 'loader' and passing it a function is a convenient way
     * of running a special loader function. @see runLoader
     *
     * @param {Helper.id} helperId
     * @param {Whatever} ...args
    */

  }, {
    key: 'run',
    value: function run(helperId) {
      var _lookup;

      var context = (0, _extends3.default)({}, this.__helperContext);

      util.hide.property(context, '__registryPath', __filename);
      util.hide.property(context, '__registryType', this.name);

      for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      if (helperId === 'loader') {
        return this.runLoader.apply(this, args);
      }

      args.push(context);

      return (_lookup = this.lookup(helperId)).run.apply(_lookup, args);
    }

    /**
     * Like run but returns a promise
    */

  }, {
    key: 'dispatch',
    value: function dispatch(helperId) {
      var _lookup2;

      var context = (0, _extends3.default)({}, this.__helperContext);

      util.hide.property(context, '__registryPath', __filename);
      util.hide.property(context, '__registryType', this.name);

      for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
        args[_key2 - 1] = arguments[_key2];
      }

      args.push(context);

      return (_lookup2 = this.lookup(helperId)).dispatch.apply(_lookup2, args);
    }

    /**
     * The Fallback registry for now will always be Skypager
     *
     * @private
    */

  }, {
    key: 'query',

    /**
     * Query this helper registry and return only helpers which match
     *
     * @example return actions that expose a CLI interface
     *
     *  skypager.actions.query((helper) => {
     *    return ('cli' in helper.definition.interfaces)
     *  })
    */
    value: function query(params) {
      return util.filterQuery(this.allHelpers(true), params);
    }

    /**
     * Lookup a helper by id
     *
     * @param {Helper.id} needle the id of the helper you want
     * @param {Boolean} strict throw an error when it is not found
     * @param {Project} fromProject only return helpers that were registered by a particular project
     *
     */

  }, {
    key: 'lookup',
    value: function lookup(needle) {
      var strict = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];
      var fromProject = arguments[2];

      var helperId = this.aliases[needle];

      if (typeof helperId === 'undefined') {
        if (this.fallback) {
          return this.fallback.lookup(needle, strict, this.host);
        }
        if (strict) {
          throw 'Could not find helper with id:' + needle;
        }
      }

      var result = this.registry[helperId];

      if (result && fromProject) {
        result.options.project = fromProject;
      }

      return result;
    }

    /**
     * Register a helper instance
     *
     * @param {Helper.id} helperId the id to reference this helper by
     * @param {Helper} helperInstance a helper object that wraps this helper file with metadata
    */

  }, {
    key: 'register',
    value: function register(helperId, helperInstance) {
      var _this2 = this;

      if (!helperInstance) {
        throw 'Error registering ' + helperId;
      }

      this.aliases[helperId] = helperId;

      if (helperInstance.name) {
        this.aliases[helperInstance.name] = helperId;
      }

      this.registry[helperId] = helperInstance;

      helperInstance.aliases && helperInstance.aliases.forEach(function (alias) {
        _this2.aliases[alias] = helperId;
      });

      return helperInstance;
    }

    /**
     * build a Helper.id for a given helper path
     *
     * @param {Path} helperURL the absolute path to this helper
     * @param {Boolean} keepExtension - keep the file extension as part of the Helper.id
    */

  }, {
    key: 'buildId',
    value: function buildId(helperURL) {
      var keepExtension = arguments.length <= 1 || arguments[1] === undefined ? false : arguments[1];

      var reg = new RegExp('^' + this.root + '/', 'i');

      if (!helperURL.match(reg)) {
        helperURL = (0, _path.basename)(helperURL);
      }

      var base = helperURL.replace(reg, '');

      if (base.match(/\/index$/i)) {
        base = base.replace(/\/index$/i, '');
      }

      base = base.split('/').reverse().slice(0, 2).reverse().join('/');

      return keepExtension ? base : base.replace(/\.\w+$/i, '');
    }

    /**
     * Run a helper loader function for this registry.
     *
     * @param {Function} fn a function which is about to load helpers for us
     * @param {Object} locals an object containing variables to inject into scope
     *
     * This will run the required function in a special context
     * where certain sugar is injected into the global scope.
     *
     * These loader functions can expect to have the following in scope:
     *
     * - registry - this
     * - [helperType] - a variable named action, model, exporter, plugin, or whatever
     * - load - a function to load in a uri. this.load.bind(registry)
     *
    */

  }, {
    key: 'runLoader',
    value: function runLoader(requiredModule) {
      var locals = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

      var fn = requiredModule.default ? requiredModule.default : requiredModule;
      var registry = this;
      var host = this.host;

      locals = (0, _assign2.default)(locals, this.helper.DSL ? this.helper.DSL : {});
      locals.load = registry.load.bind(registry);

      util.noConflict(function () {
        try {
          fn.call(host, registry, host.type);
        } catch (error) {
          host && host.logger && host.error('Error running registry loader function', error);
        }
      }, locals)();
    }

    /**
     * Load a helper by its URI or Path.
     *
     * @description
     *
     * This will require the helper in a special context where certain
     * objects and functions are injected in the global scope. This makes it easier to
     * write helpers by providing them with a specific DSL based on the helper type.
     *
     * @param {String} uri the absolute path to the helper
     * @param {Helper.id} id what id to register this helper under?
     *
     * @see helpers/definitions/model.js for example
     *
     * NOTE: Need to refactor this if im going to use webpack as a build
     * since it doesnt like dynamic require.
    */

  }, {
    key: 'load',
    value: function load(uri, id) {
      if (typeof uri !== 'string') {
        return this.loadModule.apply(this, arguments);
      }

      return this.loadPath.apply(this, arguments);
    }
  }, {
    key: 'loadModule',
    value: function loadModule(required) {
      var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

      if (typeof required === 'string') {
        return this.loadPath.apply(this, arguments);
      }

      var id = options.id;
      var uri = options.uri;

      id = id || this.buildId(uri);

      var owner = this;
      var HelperClass = this.helper;

      var helperInstance = undefined;
      var empty = (typeof required === 'undefined' ? 'undefined' : (0, _typeof3.default)(required)) === 'object' && (0, _keys2.default)(required).length === 0;
      var definition = this.helper.Definition && this.helper.Definition.current();

      if (definition) {
        helperInstance = HelperClass.fromDefinition(uri, definition, { owner: owner, id: id, required: required });
      } else {
        helperInstance = new HelperClass(uri, { owner: owner, id: id, required: required });
      }

      this.register(id, helperInstance);

      // todo: should just be a method on definition
      if (this.helper.Definition && this.helper.Definition.clearDefinition) {
        this.helper.Definition.clearDefinition();
      }

      return helperInstance;
    }
  }, {
    key: 'loadPath',
    value: function loadPath(uri, id) {
      var HelperClass = this.helper;

      var owner = this;

      id = id || this.buildId(uri);

      var helperInstance = undefined;

      try {
        var required = require(uri);
        var empty = (typeof required === 'undefined' ? 'undefined' : (0, _typeof3.default)(required)) === 'object' && (0, _keys2.default)(required).length === 0;
        var definition = this.helper.Definition && this.helper.Definition.current();

        if (empty && definition) {
          helperInstance = HelperClass.fromDefinition(uri, definition, { owner: owner, id: id, required: required });
        } else if (definition) {
          helperInstance = HelperClass.fromDefinition(uri, definition, { owner: owner, id: id, required: required });
        } else {
          helperInstance = new HelperClass(uri, { owner: owner, id: id, definition: definition, required: required });
        }

        if (!helperInstance) {
          throw 'Uh oh';
        }

        id = id.replace(/\/index$/i, '');

        this.register(id, helperInstance);

        // todo: should just be a method on definition
        if (this.helper.Definition && this.helper.Definition.clearDefinition) {
          this.helper.Definition.clearDefinition();
        }

        return helperInstance;
      } catch (error) {
        console.log('Error loading: ' + uri, error.message);
        console.log(error.stack);
      }
    }
  }, {
    key: 'filter',
    value: function filter() {
      var _allHelpers;

      return (_allHelpers = this.allHelpers(true)).filter.apply(_allHelpers, arguments);
    }
  }, {
    key: 'map',
    value: function map() {
      var _allHelpers2;

      return (_allHelpers2 = this.allHelpers(true)).map.apply(_allHelpers2, arguments);
    }
  }, {
    key: 'forEach',
    value: function forEach() {
      var _allHelpers3;

      return (_allHelpers3 = this.allHelpers(true)).forEach.apply(_allHelpers3, arguments);
    }
  }, {
    key: 'allHelpers',
    value: function allHelpers() {
      var includeFallback = arguments.length <= 0 || arguments[0] === undefined ? true : arguments[0];

      var mine = util.values(this.registry);

      if (this.fallback && includeFallback) {
        return mine.concat(this.fallback.all);
      }

      return mine;
    }
  }, {
    key: 'fallback',
    get: function get() {
      if (this.host.type !== 'framework') {
        return Fallbacks[this.name];
      }
    }
  }, {
    key: 'all',
    get: function get() {
      return this.allHelpers(false);
    }
  }, {
    key: 'available',
    get: function get() {
      var keys = (0, _keys2.default)(this.registry);

      if (this.fallback) {
        return keys.concat(this.fallback.available);
      }

      return keys;
    }
  }]);
  return Registry;
})();

var _CACHE = {};

var Builder = (function () {
  function Builder(cache) {
    (0, _classCallCheck3.default)(this, Builder);

    this.cache = cache || {};
  }

  (0, _createClass3.default)(Builder, [{
    key: 'build',
    value: function build() {
      return Registry.build.apply(Registry, arguments);
    }
  }, {
    key: 'buildAll',
    value: function buildAll(host, helpers) {
      var options = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

      var root = options.root || host.root;
      var registryCacheKey = host.registryId;

      var c = this.cache[registryCacheKey] = this.cache[registryCacheKey] || {};

      (0, _keys2.default)(helpers).filter(function (type) {
        return type !== 'default';
      }).forEach(function (type) {
        var name = util.tableize(type);

        c[name] = c[name] || Registry.build(host, helpers[type], { name: name, root: host.root });
      });

      return c;
    }
  }]);
  return Builder;
})();

module.exports = new Builder(_CACHE);

function buildAtInterface(host) {
  var chain = function chain(id) {
    return host.lookup(id);
  };

  var expand = host.loaded;

  Object.defineProperty(host, 'at', {
    configurable: true,
    enumerable: false,
    value: chain
  });

  var idPaths = host.available.concat([]);

  if (expand) {
    var expanded = idPaths.map(function (idPath) {
      return idPath.split('/');
    }).sort(function (a, b) {
      return a.length > b.length;
    });

    expanded.forEach(function (parts) {
      var id = parts.join('/');
      var first = parts[0];

      if (parts.length === 1) {
        var _util$assign, _mutatorMap;

        util.assign(chain, (_util$assign = {}, _mutatorMap = {}, _mutatorMap[first] = _mutatorMap[first] || {}, _mutatorMap[first].get = function () {
          return host.lookup(id);
        }, (0, _defineEnumerableProperties3.default)(_util$assign, _mutatorMap), _util$assign));
      }

      if (parts.length > 1) {
        (function () {
          var _util$assign2, _mutatorMap2;

          var getter = parts.pop();
          var idPath = parts.join('.').replace(/-/g, '_');
          var target = (0, _lodash.get)(chain, idPath) || {};

          util.assign(target, (_util$assign2 = {}, _mutatorMap2 = {}, _mutatorMap2[getter] = _mutatorMap2[getter] || {}, _mutatorMap2[getter].get = function () {
            return host.lookup(id);
          }, (0, _defineEnumerableProperties3.default)(_util$assign2, _mutatorMap2), _util$assign2));

          (0, _lodash.set)(chain, idPath, target);
        })();
      }
    });
  }
}